# 25 "cil/src/logic/logic_preprocess.mll"
 
  open Lexing
  type state = NORMAL | SLASH | INCOMMENT
  type end_of_buffer = NEWLINE | SPACE | CHAR
  let buf = Buffer.create 1024
  let macros = Buffer.create 1024
  let beg_of_line = Buffer.create 8
  let blacklisted_macros = [ "__STDC__"; "__STDC_HOSTED__"; "assert"]
  let is_newline = ref CHAR
  let curr_file = ref ""
  let curr_line = ref 1
  let is_ghost = ref false
  let begin_annot_line = ref 1

  let reset () =
    Buffer.clear buf;
    Buffer.clear macros;
    Buffer.clear beg_of_line;
    is_newline := CHAR;
    curr_file := "";
    curr_line := 1;
    is_ghost := false;
    begin_annot_line := 1

  let backslash = "__BACKSLASH__"

  let abort_preprocess reason outfile =
    let source = {Lexing.dummy_pos with Lexing.pos_fname = !curr_file;
                  pos_lnum = !curr_line;}
    in
    Kernel.error ~source
      "Can't preprocess annotation: %s\nAnnotation will be kept as is"
      reason;
    Buffer.output_buffer outfile buf

  let preprocess_annot suffix cpp outfile =
    (*Printf.printf "Preprocessing annotation:\n%!";
    Buffer.output_buffer stdout buf;
    print_newline(); *)
    let debug =
      Kernel.debug_atleast 3 ||
      Kernel.Debug_category.exists (fun x -> x = "parser")
    in
    let ppname =
      try Extlib.temp_file_cleanup_at_exit ~debug "ppannot" suffix
      with Extlib.Temp_file_error s ->
        Kernel.abort
          "Could not open temporary file for logic pre-processing: %s" s
    in
    let ppfile = open_out ppname in
    Buffer.output_buffer ppfile macros;
    (* NB: the three extra spaces replace the beginning of the annotation
       in order to keep the columns count accurate (at least until there's
       a macro expansion).
    *)
    Printf.fprintf ppfile "# %d %s \n   " !begin_annot_line !curr_file;
    Buffer.output_buffer ppfile beg_of_line;
    Buffer.output_buffer ppfile buf;
    (* cpp complains if the temp file does not end with a newline *)
    Buffer.clear beg_of_line;
    if not (!is_newline = NEWLINE) then output_char ppfile '\n';
    close_out ppfile;
    let cppname = Extlib.temp_file_cleanup_at_exit ~debug "cppannot" suffix in
    let res = Sys.command (cpp ppname cppname) in
    if not debug then Extlib.safe_remove ppname;
    output_string outfile "/*@";
    if !is_ghost then output_string outfile " ghost\n";
    if res <> 0 then begin
      abort_preprocess "Preprocessor call exited with an error" outfile;
      if not debug then Extlib.safe_remove cppname
    end else begin
    try
      let tmp = open_in_bin cppname in
      let tmp_buf = Buffer.create 1024 in
      Buffer.clear tmp_buf;
      let x = ref (input_char tmp) in
      let state = ref NORMAL in
      (try
         while true do
           (* we have to remove the spurious \n at the end of buffer*)
           let c = input_char tmp in
           (match !x with
                '/' ->
                  (match !state with
                       NORMAL -> state:=SLASH
                     | SLASH ->state:=INCOMMENT
                     | INCOMMENT -> ()
                  )
              | '\n' -> state:=NORMAL
              | _ -> (match !state with
                          SLASH->state:=NORMAL
                        | NORMAL | INCOMMENT -> ())
           );
           Buffer.add_char tmp_buf !x;
           x:=c;
         done;
         assert false
       with
           End_of_file ->
             if !is_newline <> CHAR
             then Buffer.add_char tmp_buf !x;
             (* one-line annotations get a new line anyway. *)
             if !state = INCOMMENT then
             Buffer.add_char tmp_buf '\n';
             let res = Buffer.contents tmp_buf in
             let res =
               Str.global_replace (Str.regexp_string backslash) "\\\\" res
             in
             (* Printf.printf "after preprocessing:\n%s%!" res; *)
             output_string outfile res;
             close_in tmp;
             if not debug then Sys.remove cppname)
    with
      | End_of_file ->
        if not debug then (try Sys.remove cppname with Sys_error _ -> ());
        abort_preprocess "Empty result in annotation pre-processing" outfile
      | Sys_error e ->
          if not debug then (try Sys.remove cppname with Sys_error _ -> ());
          abort_preprocess ("System error: " ^ e) outfile

    end;
    Printf.fprintf outfile "*/\n# %d %s\n%!" !curr_line !curr_file;
    Buffer.clear buf

  let make_newline () =
    incr curr_line;
    Buffer.clear beg_of_line

# 131 "cil/src/logic/logic_preprocess.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\247\255\248\255\249\255\250\255\251\255\001\000\004\000\
    \017\000\000\000\031\000\000\000\000\000\000\000\001\000\001\000\
    \002\000\081\000\255\255\001\000\163\000\000\000\006\000\006\000\
    \004\000\008\000\021\001\003\000\010\000\041\000\031\000\254\255\
    \032\000\033\000\253\255\252\255\082\000\251\255\253\255\254\255\
    \004\000\252\255\255\255\083\000\251\255\253\255\254\255\006\000\
    \252\255\255\255\025\001\251\255\003\000\011\000\254\255\000\000\
    \006\000\003\000\003\000\253\255\252\255\195\000\252\255\016\000\
    \254\255\014\000\012\000\009\000\009\000\253\255\052\001\247\255\
    \248\255\249\255\250\255\251\255\252\255\015\000\254\255\017\000\
    \255\255\253\255\085\000\252\255\253\255\029\000\255\255\254\255\
    \133\000\250\255\251\255\050\000\254\255\255\255\252\255\253\255\
    \026\001\251\255\252\255\043\000\254\255\255\255\253\255\130\000\
    \252\255\253\255\254\255\255\255\029\001\250\255\251\255\252\255\
    \253\255\254\255\255\255\087\000\254\255\255\255";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\008\000\008\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \004\000\255\255\255\255\255\255\255\255\255\255\255\255\004\000\
    \255\255\255\255\255\255\255\255\004\000\004\000\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\003\000\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\008\000\255\255\008\000\
    \255\255\255\255\255\255\255\255\255\255\003\000\255\255\255\255\
    \255\255\255\255\255\255\005\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\004\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\019\000\000\000\019\000\019\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\029\000\255\255\029\000\030\000\000\000\
    \035\000\034\000\000\000\000\000\037\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\044\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\051\000\000\000\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\000\000\000\000\062\000\000\000\255\255\
    \000\000\255\255\255\255\255\255\255\255\000\000\071\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\255\255\
    \000\000\000\000\083\000\000\000\000\000\255\255\000\000\000\000\
    \089\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \097\000\000\000\000\000\255\255\000\000\000\000\000\000\104\000\
    \000\000\000\000\000\000\000\000\109\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\116\000\000\000\000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\055\000\004\000\018\000\000\000\010\000\000\000\000\000\
    \000\000\000\000\000\000\028\000\000\000\000\000\000\000\065\000\
    \000\000\000\000\028\000\000\000\000\000\000\000\000\000\000\000\
    \055\000\000\000\003\000\007\000\010\000\255\255\042\000\002\000\
    \010\000\031\000\028\000\033\000\027\000\049\000\065\000\006\000\
    \032\000\028\000\060\000\027\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\081\000\010\000\
    \080\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\030\000\087\000\102\000\000\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\095\000\017\000\018\000\038\000\045\000\000\000\086\000\
    \041\000\117\000\048\000\000\000\022\000\013\000\014\000\017\000\
    \012\000\024\000\015\000\023\000\017\000\026\000\021\000\016\000\
    \009\000\017\000\025\000\056\000\039\000\057\000\058\000\059\000\
    \066\000\011\000\046\000\067\000\068\000\069\000\000\000\085\000\
    \000\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\009\000\106\000\000\000\094\000\093\000\
    \000\000\000\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\092\000\018\000\040\000\047\000\
    \020\000\107\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\065\000\064\000\000\000\000\000\
    \000\000\000\000\000\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\000\000\000\000\000\000\
    \000\000\091\000\000\000\065\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\000\000\000\000\
    \005\000\255\255\020\000\255\255\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\026\000\255\255\
    \255\255\255\255\055\000\054\000\101\000\000\000\000\000\114\000\
    \000\000\255\255\063\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\026\000\000\000\000\000\
    \000\000\055\000\000\000\000\000\100\000\000\000\078\000\110\000\
    \000\000\000\000\000\000\052\000\111\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\000\000\
    \000\000\255\255\255\255\255\255\075\000\084\000\072\000\117\000\
    \000\000\000\000\000\000\073\000\000\000\113\000\079\000\000\000\
    \000\000\000\000\000\000\077\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\076\000\000\000\099\000\000\000\
    \000\000\112\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \053\000\000\000\105\000\000\000\000\000\090\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \074\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\098\000\000\000\000\000\114\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\055\000\000\000\019\000\255\255\007\000\255\255\255\255\
    \255\255\255\255\255\255\028\000\255\255\255\255\255\255\065\000\
    \255\255\255\255\008\000\255\255\255\255\255\255\255\255\255\255\
    \055\000\255\255\000\000\000\000\007\000\027\000\040\000\000\000\
    \010\000\030\000\028\000\006\000\028\000\047\000\065\000\000\000\
    \006\000\008\000\052\000\008\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\077\000\010\000\
    \079\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\029\000\085\000\099\000\255\255\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\091\000\017\000\017\000\036\000\043\000\255\255\082\000\
    \040\000\115\000\047\000\255\255\021\000\012\000\013\000\016\000\
    \007\000\009\000\014\000\022\000\023\000\025\000\011\000\015\000\
    \007\000\017\000\024\000\053\000\036\000\056\000\057\000\058\000\
    \063\000\007\000\043\000\066\000\067\000\068\000\255\255\082\000\
    \255\255\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\010\000\103\000\255\255\091\000\088\000\
    \255\255\255\255\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\088\000\020\000\036\000\043\000\
    \017\000\103\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\061\000\061\000\255\255\255\255\
    \255\255\255\255\255\255\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\255\255\255\255\255\255\
    \255\255\088\000\255\255\061\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\255\255\255\255\
    \000\000\019\000\020\000\027\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\026\000\030\000\
    \032\000\033\000\050\000\050\000\096\000\255\255\255\255\108\000\
    \255\255\029\000\061\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\026\000\255\255\255\255\
    \255\255\050\000\255\255\255\255\096\000\255\255\070\000\108\000\
    \255\255\255\255\255\255\050\000\108\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\255\255\
    \255\255\017\000\036\000\043\000\070\000\082\000\070\000\115\000\
    \255\255\255\255\255\255\070\000\255\255\108\000\070\000\255\255\
    \255\255\255\255\255\255\070\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\070\000\255\255\096\000\255\255\
    \255\255\108\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \050\000\255\255\103\000\255\255\255\255\088\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \070\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\020\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\061\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\050\000\096\000\255\255\255\255\108\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\070\000";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \010\000\000\000\036\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\062\000\030\000\000\000\137\000\000\000\000\000\000\000\
    \000\000\002\000\251\000\000\000\001\000\000\000\000\000\038\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\009\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\009\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\009\000\035\000\001\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\009\000\000\000\000\000\001\000\000\000\000\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\001\000\000\000\000\000\022\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\022\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\017\000\022\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\000\000\000\000\000\000\000\000\027\000\000\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\000\000\000\000\000\000\000\000\
    \027\000\000\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\001\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\001\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\007\000\028\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\008\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \007\000\028\000\029\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\008\000\255\255\255\255\010\000\255\255\255\255\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\010\000\255\255\255\255\017\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\017\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\016\000\023\000\025\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\255\255\255\255\255\255\255\255\017\000\255\255\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\255\255\255\255\255\255\255\255\
    \020\000\255\255\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\026\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\026\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255";
  Lexing.lex_code = 
   "\255\004\255\255\006\255\005\255\255\006\255\255\005\255\006\255\
    \255\008\255\007\255\255\007\255\008\255\255\008\255\255\000\007\
    \001\008\255\009\255\255\000\004\001\005\002\006\003\009\255";
}

let rec main suffix cpp outfile lexbuf =
  lexbuf.Lexing.lex_mem <- Array.create 10 (-1) ;   __ocaml_lex_main_rec suffix cpp outfile lexbuf 0
and __ocaml_lex_main_rec suffix cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 155 "cil/src/logic/logic_preprocess.mll"
                                                                      m
# 517 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1) in
# 157 "cil/src/logic/logic_preprocess.mll"
      (
        if not (List.mem m blacklisted_macros) then
          Buffer.add_string macros (lexeme lexbuf);
	output_char outfile '\n';
        make_newline ();
        main suffix cpp outfile lexbuf
      )
# 527 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
let
# 164 "cil/src/logic/logic_preprocess.mll"
                                                       line
# 533 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 165 "cil/src/logic/logic_preprocess.mll"
                                     file
# 538 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(3) in
# 166 "cil/src/logic/logic_preprocess.mll"
    ( (try
        curr_line := (int_of_string line) -1
       with Failure "int_of_string" -> curr_line:= -1);
      if file <> "" then curr_file := file;
      output_string outfile (lexeme lexbuf);
      make_newline();
      main suffix cpp outfile lexbuf
    )
# 549 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
let
# 174 "cil/src/logic/logic_preprocess.mll"
                c
# 555 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 174 "cil/src/logic/logic_preprocess.mll"
                   (
      if c = !Clexer.annot_char then begin
        is_newline:=CHAR;
        begin_annot_line := ! curr_line;
        Buffer.clear buf;
        maybe_ghost suffix cpp outfile lexbuf
      end else begin
        output_string outfile (lexeme lexbuf);
        if c = '\n' then make_newline();
        Buffer.add_string beg_of_line "   ";
        comment suffix cpp outfile c lexbuf;
      end)
# 570 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
let
# 186 "cil/src/logic/logic_preprocess.mll"
                c
# 576 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 186 "cil/src/logic/logic_preprocess.mll"
                   (
      if c = !Clexer.annot_char then begin
        Buffer.clear buf;
        begin_annot_line := !curr_line;
        is_newline:=CHAR;
        maybe_oneline_ghost suffix cpp outfile lexbuf
      end
      else if c = '\n' then begin
        make_newline ();
        output_string outfile (lexeme lexbuf);
        main suffix cpp outfile lexbuf
      end
      else begin
        output_string outfile (lexeme lexbuf);
        oneline_comment suffix cpp outfile lexbuf;
      end)
# 595 "cil/src/logic/logic_preprocess.ml"

  | 4 ->
# 202 "cil/src/logic/logic_preprocess.mll"
         ( flush outfile )
# 600 "cil/src/logic/logic_preprocess.ml"

  | 5 ->
# 203 "cil/src/logic/logic_preprocess.mll"
         (
      make_newline ();
      output_char outfile '\n';
      main suffix cpp outfile lexbuf )
# 608 "cil/src/logic/logic_preprocess.ml"

  | 6 ->
# 207 "cil/src/logic/logic_preprocess.mll"
        ( 
      Buffer.add_char beg_of_line ' ';
      output_char outfile '"'; 
      c_string suffix cpp outfile lexbuf )
# 616 "cil/src/logic/logic_preprocess.ml"

  | 7 ->
# 211 "cil/src/logic/logic_preprocess.mll"
        ( 
      Buffer.add_char beg_of_line ' ';
      output_char outfile '\'';
      c_char suffix cpp outfile lexbuf )
# 624 "cil/src/logic/logic_preprocess.ml"

  | 8 ->
let
# 215 "cil/src/logic/logic_preprocess.mll"
         c
# 630 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 215 "cil/src/logic/logic_preprocess.mll"
           (
      Buffer.add_char beg_of_line ' ';
      output_char outfile c;
      main suffix cpp outfile lexbuf )
# 637 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_main_rec suffix cpp outfile lexbuf __ocaml_lex_state

and c_string suffix cpp outfile lexbuf =
    __ocaml_lex_c_string_rec suffix cpp outfile lexbuf 36
and __ocaml_lex_c_string_rec suffix cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 221 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_string beg_of_line "  ";
           output_string outfile (lexeme lexbuf);
           c_string suffix cpp outfile lexbuf )
# 650 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 224 "cil/src/logic/logic_preprocess.mll"
       ( Buffer.add_char beg_of_line ' '; 
         output_char outfile '"';
         main suffix cpp outfile lexbuf )
# 657 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 227 "cil/src/logic/logic_preprocess.mll"
       ( make_newline ();
         output_char outfile '\n';
         c_string suffix cpp outfile lexbuf
       )
# 665 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
# 231 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_string beg_of_line "  ";
           output_string outfile (lexeme lexbuf);
           c_string suffix cpp outfile lexbuf )
# 672 "cil/src/logic/logic_preprocess.ml"

  | 4 ->
let
# 234 "cil/src/logic/logic_preprocess.mll"
       c
# 678 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 234 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_char beg_of_line ' ';
           output_char outfile c;
           c_string suffix cpp outfile lexbuf )
# 684 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_c_string_rec suffix cpp outfile lexbuf __ocaml_lex_state

and c_char suffix cpp outfile lexbuf =
    __ocaml_lex_c_char_rec suffix cpp outfile lexbuf 43
and __ocaml_lex_c_char_rec suffix cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 239 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_string beg_of_line "  ";
           output_string outfile (lexeme lexbuf);
           c_char suffix cpp outfile lexbuf )
# 697 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 242 "cil/src/logic/logic_preprocess.mll"
      ( Buffer.add_char beg_of_line ' '; 
         output_char outfile '\'';
         main suffix cpp outfile lexbuf )
# 704 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 245 "cil/src/logic/logic_preprocess.mll"
       ( make_newline ();
         output_char outfile '\n';
         c_char suffix cpp outfile lexbuf
       )
# 712 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
# 249 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_string beg_of_line "  ";
           output_string outfile (lexeme lexbuf);
           c_char suffix cpp outfile lexbuf )
# 719 "cil/src/logic/logic_preprocess.ml"

  | 4 ->
let
# 252 "cil/src/logic/logic_preprocess.mll"
       c
# 725 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 252 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_char beg_of_line ' ';
           output_char outfile c;
           c_char suffix cpp outfile lexbuf )
# 731 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_c_char_rec suffix cpp outfile lexbuf __ocaml_lex_state

and maybe_ghost suffix cpp outfile lexbuf =
    __ocaml_lex_maybe_ghost_rec suffix cpp outfile lexbuf 50
and __ocaml_lex_maybe_ghost_rec suffix cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 257 "cil/src/logic/logic_preprocess.mll"
                 space
# 743 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 257 "cil/src/logic/logic_preprocess.mll"
                      (
     Buffer.add_string buf space;
     maybe_ghost suffix cpp outfile lexbuf )
# 749 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 260 "cil/src/logic/logic_preprocess.mll"
         (
      is_newline := NEWLINE;
      incr curr_line;
      Buffer.add_char buf '\n';
      maybe_ghost suffix cpp outfile lexbuf
    )
# 759 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 267 "cil/src/logic/logic_preprocess.mll"
      ( is_ghost := true;
        Buffer.add_string buf "     ";
        annot suffix cpp outfile lexbuf
      )
# 767 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
# 272 "cil/src/logic/logic_preprocess.mll"
         ( main suffix cpp outfile lexbuf )
# 772 "cil/src/logic/logic_preprocess.ml"

  | 4 ->
let
# 273 "cil/src/logic/logic_preprocess.mll"
         c
# 778 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 273 "cil/src/logic/logic_preprocess.mll"
           ( Buffer.add_char buf c; is_ghost:=false;
             annot suffix cpp outfile lexbuf)
# 783 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_maybe_ghost_rec suffix cpp outfile lexbuf __ocaml_lex_state

and maybe_oneline_ghost suffix cpp outfile lexbuf =
    __ocaml_lex_maybe_oneline_ghost_rec suffix cpp outfile lexbuf 61
and __ocaml_lex_maybe_oneline_ghost_rec suffix cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 276 "cil/src/logic/logic_preprocess.mll"
                 space
# 795 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 276 "cil/src/logic/logic_preprocess.mll"
                      (
     Buffer.add_string buf space;
     maybe_oneline_ghost suffix cpp outfile lexbuf )
# 801 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 279 "cil/src/logic/logic_preprocess.mll"
         (
      incr curr_line;
      main suffix cpp outfile lexbuf
    )
# 809 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 284 "cil/src/logic/logic_preprocess.mll"
      ( is_ghost := true;
        Buffer.add_string buf "     ";
        oneline_annot suffix cpp outfile lexbuf
      )
# 817 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
let
# 288 "cil/src/logic/logic_preprocess.mll"
         c
# 823 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 289 "cil/src/logic/logic_preprocess.mll"
      (
        Buffer.add_char buf c;
        is_ghost:=false;
        oneline_annot suffix cpp outfile lexbuf
      )
# 831 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_maybe_oneline_ghost_rec suffix cpp outfile lexbuf __ocaml_lex_state

and annot suffix cpp outfile lexbuf =
    __ocaml_lex_annot_rec suffix cpp outfile lexbuf 70
and __ocaml_lex_annot_rec suffix cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 295 "cil/src/logic/logic_preprocess.mll"
          ( preprocess_annot suffix cpp outfile;
            main suffix cpp outfile lexbuf )
# 843 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 297 "cil/src/logic/logic_preprocess.mll"
         ( is_newline := NEWLINE;
           incr curr_line;
           Buffer.add_char buf '\n';
           annot suffix cpp outfile lexbuf )
# 851 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 301 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_string buf "//";
           annot_comment suffix cpp outfile lexbuf )
# 857 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
# 303 "cil/src/logic/logic_preprocess.mll"
        (
      if !is_newline = NEWLINE then is_newline:=SPACE;
      Buffer.add_char buf ' ';
      annot suffix cpp outfile lexbuf )
# 865 "cil/src/logic/logic_preprocess.ml"

  | 4 ->
# 307 "cil/src/logic/logic_preprocess.mll"
         (
      if !is_newline = NEWLINE then is_newline:=SPACE;
      Buffer.add_char buf ' ';
      annot suffix cpp outfile lexbuf )
# 873 "cil/src/logic/logic_preprocess.ml"

  | 5 ->
# 315 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_string buf backslash;
           annot suffix cpp outfile lexbuf )
# 879 "cil/src/logic/logic_preprocess.ml"

  | 6 ->
# 317 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_char buf '\'';
           char suffix annot cpp outfile lexbuf )
# 885 "cil/src/logic/logic_preprocess.ml"

  | 7 ->
# 319 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_char buf '"';
           string suffix annot cpp outfile lexbuf )
# 891 "cil/src/logic/logic_preprocess.ml"

  | 8 ->
let
# 321 "cil/src/logic/logic_preprocess.mll"
         c
# 897 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 321 "cil/src/logic/logic_preprocess.mll"
           ( is_newline := CHAR;
             Buffer.add_char buf c;
             annot suffix cpp outfile lexbuf )
# 903 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_annot_rec suffix cpp outfile lexbuf __ocaml_lex_state

and annot_comment suffix cpp outfile lexbuf =
    __ocaml_lex_annot_comment_rec suffix cpp outfile lexbuf 82
and __ocaml_lex_annot_comment_rec suffix cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 326 "cil/src/logic/logic_preprocess.mll"
         ( incr curr_line; is_newline:=NEWLINE;
           Buffer.add_char buf '\n';
           annot suffix cpp outfile lexbuf
         )
# 917 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 330 "cil/src/logic/logic_preprocess.mll"
         ( preprocess_annot suffix cpp outfile;
           main suffix cpp outfile lexbuf )
# 923 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 332 "cil/src/logic/logic_preprocess.mll"
        ( abort_preprocess "eof in the middle of a comment" outfile )
# 928 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
let
# 333 "cil/src/logic/logic_preprocess.mll"
         c
# 934 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 333 "cil/src/logic/logic_preprocess.mll"
           (
    Buffer.add_char buf c;
    annot_comment suffix cpp outfile lexbuf )
# 940 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_annot_comment_rec suffix cpp outfile lexbuf __ocaml_lex_state

and char suffix annot cpp outfile lexbuf =
    __ocaml_lex_char_rec suffix annot cpp outfile lexbuf 88
and __ocaml_lex_char_rec suffix annot cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 339 "cil/src/logic/logic_preprocess.mll"
         ( incr curr_line; is_newline:=NEWLINE;
           Buffer.add_char buf '\n';
           char suffix annot cpp outfile lexbuf
         )
# 954 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 343 "cil/src/logic/logic_preprocess.mll"
         ( is_newline:=CHAR;
           Buffer.add_char buf '\'';
           annot suffix cpp outfile lexbuf )
# 961 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 346 "cil/src/logic/logic_preprocess.mll"
          ( is_newline:=CHAR;
            Buffer.add_string buf "\\'";
            char suffix annot cpp outfile lexbuf )
# 968 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
# 349 "cil/src/logic/logic_preprocess.mll"
           ( is_newline:=CHAR;
            Buffer.add_string buf "\\\\";
            char suffix annot cpp outfile lexbuf )
# 975 "cil/src/logic/logic_preprocess.ml"

  | 4 ->
# 352 "cil/src/logic/logic_preprocess.mll"
        ( abort_preprocess "eof while parsing a char literal" outfile )
# 980 "cil/src/logic/logic_preprocess.ml"

  | 5 ->
let
# 353 "cil/src/logic/logic_preprocess.mll"
         c
# 986 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 353 "cil/src/logic/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_char buf c;
             char suffix annot cpp outfile lexbuf )
# 992 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_char_rec suffix annot cpp outfile lexbuf __ocaml_lex_state

and string suffix annot cpp outfile lexbuf =
    __ocaml_lex_string_rec suffix annot cpp outfile lexbuf 96
and __ocaml_lex_string_rec suffix annot cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 358 "cil/src/logic/logic_preprocess.mll"
         ( incr curr_line; is_newline:=NEWLINE;
           Buffer.add_char buf '\n'; string suffix annot cpp outfile lexbuf
         )
# 1005 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 361 "cil/src/logic/logic_preprocess.mll"
        ( is_newline:=CHAR; Buffer.add_char buf '"';
          annot suffix cpp outfile lexbuf )
# 1011 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 363 "cil/src/logic/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_string buf "\\\"";
             string suffix annot cpp outfile lexbuf )
# 1018 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
# 366 "cil/src/logic/logic_preprocess.mll"
        ( abort_preprocess "eof while parsing a string literal" outfile )
# 1023 "cil/src/logic/logic_preprocess.ml"

  | 4 ->
let
# 367 "cil/src/logic/logic_preprocess.mll"
         c
# 1029 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 367 "cil/src/logic/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_char buf c;
             string suffix annot cpp outfile lexbuf )
# 1035 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_string_rec suffix annot cpp outfile lexbuf __ocaml_lex_state

and comment suffix cpp outfile c lexbuf =
    __ocaml_lex_comment_rec suffix cpp outfile c lexbuf 103
and __ocaml_lex_comment_rec suffix cpp outfile c lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 373 "cil/src/logic/logic_preprocess.mll"
        (
      Buffer.add_char beg_of_line ' ';
      output_string outfile (lexeme lexbuf);
      if c = '*' then
        main suffix cpp outfile lexbuf
      else
        comment suffix cpp outfile '/' lexbuf
      )
# 1053 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 381 "cil/src/logic/logic_preprocess.mll"
         ( make_newline (); output_char outfile '\n';
           comment suffix cpp outfile '\n' lexbuf )
# 1059 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 383 "cil/src/logic/logic_preprocess.mll"
        ( abort_preprocess "eof while parsing C comment" outfile)
# 1064 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
let
# 384 "cil/src/logic/logic_preprocess.mll"
         c
# 1070 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 384 "cil/src/logic/logic_preprocess.mll"
           (
      Buffer.add_char beg_of_line ' ';
      output_char outfile c;
      comment suffix cpp outfile c lexbuf)
# 1077 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_comment_rec suffix cpp outfile c lexbuf __ocaml_lex_state

and oneline_annot suffix cpp outfile lexbuf =
    __ocaml_lex_oneline_annot_rec suffix cpp outfile lexbuf 108
and __ocaml_lex_oneline_annot_rec suffix cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 390 "cil/src/logic/logic_preprocess.mll"
             (
      incr curr_line;
      preprocess_annot suffix cpp outfile;
      main suffix cpp outfile lexbuf )
# 1091 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
# 394 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_char buf ' ';
           oneline_annot suffix cpp outfile lexbuf
         )
# 1098 "cil/src/logic/logic_preprocess.ml"

  | 2 ->
# 397 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_string buf backslash;
           oneline_annot suffix cpp outfile lexbuf )
# 1104 "cil/src/logic/logic_preprocess.ml"

  | 3 ->
# 399 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_char buf '\'';
           char suffix oneline_annot cpp outfile lexbuf )
# 1110 "cil/src/logic/logic_preprocess.ml"

  | 4 ->
# 401 "cil/src/logic/logic_preprocess.mll"
         ( Buffer.add_char buf '"';
           string suffix oneline_annot cpp outfile lexbuf )
# 1116 "cil/src/logic/logic_preprocess.ml"

  | 5 ->
let
# 403 "cil/src/logic/logic_preprocess.mll"
         c
# 1122 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 403 "cil/src/logic/logic_preprocess.mll"
           ( Buffer.add_char buf c;
             oneline_annot suffix cpp outfile lexbuf )
# 1127 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_oneline_annot_rec suffix cpp outfile lexbuf __ocaml_lex_state

and oneline_comment suffix cpp outfile lexbuf =
    __ocaml_lex_oneline_comment_rec suffix cpp outfile lexbuf 115
and __ocaml_lex_oneline_comment_rec suffix cpp outfile lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 409 "cil/src/logic/logic_preprocess.mll"
      ( make_newline();
        output_string outfile (lexeme lexbuf);
        main suffix cpp outfile lexbuf)
# 1140 "cil/src/logic/logic_preprocess.ml"

  | 1 ->
let
# 412 "cil/src/logic/logic_preprocess.mll"
         c
# 1146 "cil/src/logic/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 412 "cil/src/logic/logic_preprocess.mll"
           ( output_char outfile c;
             oneline_comment suffix cpp outfile lexbuf)
# 1151 "cil/src/logic/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_oneline_comment_rec suffix cpp outfile lexbuf __ocaml_lex_state

;;

# 415 "cil/src/logic/logic_preprocess.mll"
 
  let file suffix cpp filename =
    reset ();
    let debug = Kernel.Debug_category.exists (fun x -> x = "parser") in
    let inchan = open_in_bin filename in
    let lex = Lexing.from_channel inchan in
    let ppname =
      Extlib.temp_file_cleanup_at_exit ~debug
        (Filename.basename filename) ".pp"
    in
    let ppfile = open_out ppname in
    main suffix cpp ppfile lex;
    close_in inchan;
    close_out ppfile;
    ppname

# 1174 "cil/src/logic/logic_preprocess.ml"
